import numpy as np

def manifold_dim():
    return (2,2)

def vector_dim():
    return (1,)

def SVDO(so2_mat):
    assert so2_mat.shape == (2,2)
    U,S,VT = np.linalg.svd(so2_mat)
    S_dash = np.identity(2)
    S_dash[1,1] = np.linalg.det(U@VT)
    return U@S_dash@VT


def inv(so2_mat):
    assert se2_mat.shape == (2,2)
    return se2_mat.T

def Exp(theta):
    cosine = np.cos(theta)
    sine = np.sin(theta)
    so2_mat = np.array([[cosine, -sine],[sine, cosine]])
    return so2_mat

def Log(so2_mat):
    cosine = so2_mat[0,0]
    sine = so2_mat[1,0]
    theta = np.arctan2(sine,cosine)
    return theta

def adjoint(so2_mat):
    return 1.0




from . import SO3
from .utils import *

def manifold_dim():
    return (4,4)

def vector_dim():
    return (6,)

def to_Rt(se3_mat):
    assert se3_mat.shape == (4,4)
    R = se3_mat[0:3,0:3]
    t = se3_mat[0:3,3]
    return R, t

def from_Rt(R,t):
    t = t.flatten()
    assert R.shape == (3,3)
    assert t.shape == (3,)
    T = np.identity(4)
    T[:3,:3] = R
    T[:3,3] = t
    return T

def Rx(angle, t=[0,0,0], unit='rad'):
    t = np.array(t).flatten()
    assert t.shape == (3,)
    R = SO3.Rx(angle, unit)
    return from_Rt(R,t)




def SVDO(se3_mat):
    R,t = to_Rt(se3_mat)
    R = SO3.SVDO(R)
    return from_Rt(R,t)


def wedge(vec):
    vec = np.array(vec)
    assert vec.shape == (6,)
    SE3_lie_alg = np.zeros((4,4))
    rho = vec[0:3]
    theta = vec[3:6]
    SO3_lie_alg = SO3.wedge(theta)
    SE3_lie_alg[0:3,0:3] = SO3_lie_alg
    SE3_lie_alg[0:3,3] = rho
    return SE3_lie_alg

def vee(lie_alg):
    assert lie_alg.shape == (4,4)
    SO3_lie_alg = lie_alg[0:3,0:3]
    rho = lie_alg[0:3,3]
    omega = SO3.vee(SO3_lie_alg)
    return np.concatenate((rho,omega))

def Exp(vec):
    vec = np.array(vec)
    assert vec.shape == (6,)
    transf_mat = np.identity(4)
    rho = vec[0:3]
    theta = vec[3:6]
    rot_mat = SO3.Exp(theta)
    transf_mat[0:3,0:3] = rot_mat
    transf_mat[0:3,3] = SO3.left_jacob(theta) @ rho
    return transf_mat

def exp(lie_alg):
    assert lie_alg.shape == (4,4)
    transf_mat = np.identity(4)
    u_skew = lie_alg[:3,:3]
    rho = lie_alg[:3, 3]
    R = SO3.exp(u_skew)
    transf_mat[0:3,0:3] = R
    transf_mat[0:3,3] = SO3.left_jacob(theta) @ rho
    return transf_mat

def Log(se3_mat):
    assert se3_mat.shape == (4,4)
    R,t = to_Rt(se3_mat)
    theta = SO3.Log(R)
    rho = SO3.inv_left_jacob(theta) @ t
    rho = np.squeeze(rho)
    return np.concatenate((rho,theta))

def log(se3_mat):
    return exp(Log(se3_mat))

def adjoint(se3_mat):
    ad = np.zeros((6,6))
    rot_mat, transl = to_Rt(se3_mat)
    skew_t = SO3.wedge(transl)
    ad[:3,:3] = rot_mat
    ad[:3, 3:6] = skew_t@rot_mat
    ad[3:6, 3:6] = rot_mat
    return ad



def jacob_q_term(vec):
    assert vec.shape == (6,)
    rho = vec[:3] # rho
    th = vec[3:] # theta
    an = np.linalg.norm(th) # angle
    skew_om = skew3(th)
    skew_rho = skew3(rho)
    cos_an = np.cos(an)
    sin_an = np.sin(an)

    term1 = 0.5*skew_rho
    term2 = (an - sin_an)/an**3
    term3 = skew_om @ skew_rho + skew_rho @ skew_om + skew_om @ skew_rho @ skew_om
    term4 = (1-an**2/2-np.cos(an))/an**4
    term5 = skew_om@skew_om@skew_rho+skew_rho@skew_om@skew_om-3*skew_om@skew_rho@skew_om
    term6 = 0.5*((1-an**2/2-cos_an)/(an**4)-3*(an-sin_an-an**3/6)/(an**5))
    term7 = skew_om@skew_rho@skew_om@skew_om+skew_om@skew_om@skew_rho@skew_om
    Q = term1 + term2*term3 - term4*term5 - term6*term7
    return Q


def left_jacobian(vec):
    vec = np.array(vec)
    assert vec.shape == (6,)
    if np.isclose(np.linalg.norm(vec), 0.):
        return np.identity(6) #+ 0.5 * wedge(vec)
    omega = vec[3:6]

    jacob = np.zeros((6,6))
    SO3_left_jacobian = SO3.left_jacob(omega)
    Q = jacob_q_term(vec)
    jacob[0:3,0:3] = SO3_left_jacobian
    jacob[0:3,3:] = Q
    jacob[3:,3:] = SO3_left_jacobian
    return jacob


def inv_left_jacobian(vec):
    vec = np.array(vec)
    assert vec.shape == (6,)
    if np.isclose(np.linalg.norm(vec), 0.):
        return np.identity(6) # todo: add first order taylor exp
    inv_jacob = np.zeros((6,6))
    theta_bold = vec[3:6]
    SO3_inv_left_jacobian = SO3.inv_left_jacob(theta_bold)
    Q = jacob_q_term(vec)
    inv_jacob[0:3,0:3] = SO3_inv_left_jacobian
    inv_jacob[0:3,3:] = -SO3_inv_left_jacobian@Q@SO3_inv_left_jacobian
    inv_jacob[3:,3:] = SO3_inv_left_jacobian
    return inv_jacob

def right_jacobian(vec):
    return left_jacobian(-vec)

def inv_right_jacobian(vec):
    vec = np.array(vec)
    assert vec.shape == (6,)
    return inv_left_jacobian(-vec)





from .utils import *
import numpy as np

def inv(rot_mat):
    return rot_mat.T

def Rx(angle, unit='rad'):
    angle = angle_as_rad(angle, unit)
    vec = np.array([angle, .0, .0])
    return Exp(vec)

def Ry(angle, unit='rad'):
    angle = angle_as_rad(angle, unit)
    vec = np.array([.0, angle, .0])
    return Exp(vec)

def Rz(angle, unit='rad'):
    angle = angle_as_rad(angle, unit)
    vec = np.array([.0, .0, angle])
    return Exp(vec)

def SVDO(R):
    assert R.shape == (3,3)
    U,S,VT = np.linalg.svd(R)
    S_dash = np.identity(3)
    S_dash[2,2] = np.linalg.det(U@VT)
    return U@S_dash@VT

def wedge(vec):
    vec = np.squeeze(vec)
    assert vec.shape == (3,)
    return skew3(vec)

def vee(lie_alg):
    assert lie_alg.shape == (3,3)
    vec = np.zeros((3))
    vec[0] = lie_alg[2,1]
    vec[1] = lie_alg[0,2]
    vec[2] = lie_alg[1,0]
    return vec

def log(rot_mat):
    cosine = np.clip((np.trace(rot_mat) - 1.0)/2.0, -1.0, 1.0)
    angle = np.arccos(cosine)
    if np.isclose(angle, 0.):
        return rot_mat - np.identity(3)
    lie_alg = (angle/(2*np.sin(angle)))*(rot_mat-rot_mat.transpose())
    return lie_alg

def Log(rot_mat):
    assert rot_mat.shape == (3,3)
    cos_theta = (np.trace(rot_mat) - 1) / 2
    cos_upper_lim = 1.0
    cos_lower_lim = -1.0
    cos_theta = np.clip(cos_theta, cos_lower_lim, cos_upper_lim)
    theta = np.arccos(cos_theta)
    if np.isclose(theta, 0.0):
        return vee(rot_mat - np.eye(3))
    return (theta / (2 * np.sin(theta))) * vee(rot_mat - rot_mat.T)


def exp(vec):
    assert lie_alg.shape == (3,3)
    return Exp(vee(vec))

def Exp(vec):
    vec = np.squeeze(vec)
    assert vec.shape == (3,)
    angle = np.linalg.norm(vec)
    if np.isclose(angle, 0.):
        return np.identity(3)+wedge(vec)
    axis = vec/angle
    return rodriguez(wedge(axis), angle)

def adjoint(rot_mat):
    return rot_mat

def left_jacob(vec):
    theta = np.linalg.norm(vec)
    if np.isclose(theta, 0.0):
        return np.eye(3) + 0.5 * wedge(vec)
    theta_bold_skew = skew3(vec)
    term1 = ((theta - np.sin(theta)) / theta**3) * theta_bold_skew @ theta_bold_skew
    term2 = np.eye(3) + ((1 - np.cos(theta)) / theta**2) * theta_bold_skew
    return term1 + term2


def right_jacob(vec):
    left_jacob = left_jacob(vec)
    right_jacob = left_jacob.T
    return right_jacob


def inv_left_jacob(vec):
    theta = np.linalg.norm(vec)
    if np.isclose(theta, 0.0):
        return np.eye(3) - 0.5 * wedge(vec)
    theta_bold_skew = skew3(vec)
    fraq_1 = 1 / theta**2
    fraq_2 = (1 + np.cos(theta)) / (2 * theta * np.sin(theta))
    term1 = np.eye(3) - 0.5 * theta_bold_skew
    term2 = (fraq_1 - fraq_2) * theta_bold_skew @ theta_bold_skew
    return  term1 + term2

def inv_right_jacob(vec):
    inv_left_jacob = inv_left_jacob(vec)
    inv_right_jacob = inv_left_jacob.T
    return inv_right_jacob

